;;;; cl-scopes/csys - concurrent cybernetic communication systems

(defpackage :scopes/csys
  (:use :common-lisp)
  (:local-nicknames (:actor :scopes/core/actor)
                    (:async :scopes/util/async)
                    (:config :scopes/config)
                    (:core :scopes/core)
                    (:message :scopes/core/message)
                    (:shape :scopes/shape)
                    (:util :scopes/util)
                    (:alx :alexandria))
  (:export #:*dispatcher* #:setup #:start #:shutdown
           #:printer
           ))

(in-package :scopes/csys)

(defvar *dispatcher* nil)

(defun setup (&optional (cfg config:*root*))
  (setf *dispatcher* (make-instance 'core:context :config cfg))
  (setf (core:mailbox *dispatcher*) (async:make-mb))
  (dolist (a (config:actions cfg))
          (core:add-action *dispatcher* (car a) (cadr a))))

(defun start ()
  (async:init)
  (actor:start (core:mailbox *dispatcher*) 
               (lambda (msg) (funcall #'handle-message msg)) 
               :foreground t))

(defun handle-message (msg)
  (let ((hdlrs (core:select msg (core:actions *dispatcher*))))
    (if hdlrs
      (mapcar #'(lambda (hdlr) (run-action hdlr msg)) hdlrs)
      (util:lgw "no action selected" msg)))
  nil)

(defun run-action (job msg)
  (let ((ch (async:make-ch)))
    (async:submit-task ch (lambda () (funcall job msg)))))

(defun shutdown (msg &optional (delay 0.1))
  (sleep delay)
  (format t "~&*** shutdown ~a~%" msg)
  (actor:stop (core:mailbox *dispatcher*)))

;;;; example behaviors / actions

(defun printer (msg)
  (format t "~&*** message: ~a~%" msg))
